;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - files.ss                                                       #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

; this module contains the file writing and reading functioins

(module mreddesigner-file mzscheme
  (require (lib "class.ss")
           (lib "mred.ss" "mred")
           (lib "list.ss")
           (lib "pretty.ss")
           (file "mreddesigner-misc.ss")
           (file "mreddesigner-widgets.ss")
           (file "mreddesigner-preview.ss")
           (file "mreddesigner-default-values.ss")
  )
  
  ; this list contains all properties which should be removed before writing out
  ; FIXME: callback should not be removed, but for the time being
  (define props-to-remove-write '(icon pointer callback paint-callback))
  (define (widget-write lst)
    (cond
      ((null? lst) '())
      ((list? (car lst))
       (let*
         ((full  (send hierarchy-widget get-item (caar lst)))
          (type  (cadr (assoc 'type full)))
          (props (assoc-remove full props-to-remove-write))
          (head  (append (list type (list 'id (caar lst))) props))
         )
         (cons (append head (list (cons 'children (widget-write (cdar lst)))))
               (widget-write (cdr lst))
         )
       )
      )
      (else
        (let*
          ((full  (send hierarchy-widget get-item (car lst)))
           (type  (cadr (assoc 'type full)))
           (props (assoc-remove full props-to-remove-write))
          )
          (cons (append (list type (list 'id (car lst))) props) (widget-write (cdr lst)))
        )
      )
    )
  )
  
  (provide mreddesigner-write)
  (define (mreddesigner-write file-name)
    (call-with-output-file file-name
      (lambda (p-out)
        (fprintf p-out "~s\n" '(*mred-file-2*))
        (pretty-print (widget-write (send hierarchy-widget get-hierarchy)) p-out))
      'replace
    )
  )
  
  ; --------------------------------------------------------------------------------------
  
  ; this list contains all properties which should be removed before adding to hierarchy
  ; FIXME: callback should not be removed, but for the time being
  (define props-to-remove-read '(id callback children paint-callback))
  
  (define (widget-parse lst parent)
    (let loop ((lst lst))
      (if (null? lst)
        #f
        (let*
          ((head (car lst))
           ;
           (type     (car head))
           (full     (cdr head))
           (id       (cadr (assoc 'id full)))
           (children (assoc 'children full))
           (callback (assoc 'callback full))
           (item     (assoc-remove full props-to-remove-read))
           (error    #f)
          )
          ; this is a new property, it was not in version 2.0
          (if (not (assoc 'hidden item))
            (set! item (append item '((hidden #f))))
          )
          (set! item (append (list (list 'callback (lambda (w e) (void))))
                             (if (equal? type 'canvas)
                               (list (list 'paint-callback (lambda (w e) (void))))
                               '()
                             )
                             item))
          ; before we add to hierarchy, check that it has all properties that we need
          (for-each
            (lambda (prop)
              (if (not (assoc (car prop) item))
                (begin
                  (set! error #t)
                  (message-box "Error" 
                               (string-append "Property '"
                                              (symbol->string (car prop))
                                              "' is missing from '"
                                              id
                                              "' "
                               )
                               #f '(ok stop))
                )
              )
            )
            (widget-default-values type)
          )
          (if (not error)
            (begin
              ; add widget to hierarchy list
              (send hierarchy-widget
                    item-append parent id item
                    (icon-get type)
                    ; it is a container if it has children or if it is the project (root)
                    (if (or (equal? type 'project) children) #t #f)
                    #f)
              ; if there are children add them as well
              (if children
                (widget-parse (cdr children) id)
              )
              (loop (cdr lst))
            )
          )
        )
      )
    )
  )
  
  
  ; it returns #t if everything went well, otherwise #f
  (provide mreddesigner-read)
  (define (mreddesigner-read file-name)
    (call-with-input-file file-name
      (lambda (p-in)
        (if (equal? (read p-in) '(*mred-file-2*))
          (let 
            ((old-lst (send hierarchy-widget get-hierarchy))
             (new-lst 
               (with-handlers 
                 ((exn:fail? (lambda (exn)
                               (message-box "Error" 
                                            (string-append "Error while reading MrEd file: "
                                                           (exn-message exn))
                                            #f '(ok stop))
                               #f)))
                 (read p-in)))
             (ok? #f)
            )
            ; hide current preview
            (preview-show-frames old-lst #f)
            ; clear the hierarchy widget
            (send hierarchy-widget clear)
            (if new-lst
              (begin
                (set! ok? #t)
                ; read in and add widgets to hierarchy
                (send hierarchy-widget update-view! #f)
                (widget-parse new-lst #f)
                (send hierarchy-widget update-view! #t)
                ; generate a new preview
                (preview-update-all)
                ; show the preview
                (preview-show-frames (send hierarchy-widget get-hierarchy) #t)
              )
            )
            ; notify properties window
            (send properties-panel empty)
            ;
            (changed! #t)
            ok?
          )
          (begin
            (message-box "Error" "The file format is not valid." #f '(ok stop))
            #f)))))


  
  
  
) ; end of module

