;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-func.ss                                              #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-hierarchy-func mzscheme
  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (file "mreddesigner-misc.ss")
           (file "mreddesigner-widgets.ss")
           (file "mreddesigner-preview.ss")
           (file "mreddesigner-default-values.ss")
           (file "mreddesigner-callback.ss")
  )
  
  ; it returns #f if it the movement is not allowed
  ; otherwise it returns a 5 element list
  (define (move-up-cond lst id result)
    (let*
      ((obj     (list-ref result 0))
       (pparent (list-ref result 1))
       (parent  (list-ref result 2))
       (prev    (list-ref result 3))
       (first?  (list-ref result 4))
       (obj-type (if (list? obj)
                   (send hierarchy-widget item-property-get (car obj) 'type #f)
                   (send hierarchy-widget item-property-get obj 'type #f)
                 )
       )
       (parent-type (if (list? parent)
                      (send hierarchy-widget item-property-get (car parent) 'type #f)
                      (send hierarchy-widget item-property-get parent 'type #f)
                    )
       )
       (prev-type (if prev
                    (if (list? prev)
                      (send hierarchy-widget item-property-get (car prev) 'type #f)
                      (send hierarchy-widget item-property-get prev 'type #f)
                    )
                    #f
                  )
       )
      )
      ; (car lst) == "project"
      ; IF
      ; - there is a parent of the parent and
      ;   it is not true that it is the first element and pparent == project
      ; - it is not a tab (a tab cannot be moved out of a tab-panel upwards)
      ; - the previous element is not a tab-panel 
      ;   (an element cannot be moved upwards if it would move into a tab-panel)
      ; - it it not true that the element is the first and its parent is a tab
      ;   (the first element of a tab cannot moved out upwards)
      (if (and (and pparent (not (and first? (equal? pparent (car lst)))))
               (not (equal? obj-type 'tab))
               (not (equal? prev-type 'tab-panel))
               (not (and first? (equal? parent-type 'tab)))
          )
        result 
        #f
      )
    )
  )
  
  ; a project contains frames
  ; frames cannot move into another
  ; cannot move something out of a frame
  (define (move-down-cond lst id result)
    (let*
      ((obj     (list-ref result 0))
       (pparent (list-ref result 1))
       (parent  (list-ref result 2))
       (next    (list-ref result 3))
       (last?   (list-ref result 4))
       (obj-type (if (list? obj)
                   (send hierarchy-widget item-property-get (car obj) 'type #f)
                   (send hierarchy-widget item-property-get obj 'type #f)
                 )
       )
       (parent-type (if (list? parent)
                      (send hierarchy-widget item-property-get (car parent) 'type #f)
                      (send hierarchy-widget item-property-get parent 'type #f)
                    )
       )
       (next-type (if next
                    (if (list? next)
                      (send hierarchy-widget item-property-get (car next) 'type #f)
                      (send hierarchy-widget item-property-get next 'type #f)
                    )
                    #f
                  )
       )
      )
      (if (and (and pparent (not (and last? (equal? pparent (car lst)))))
               (not (equal? obj-type 'tab))
               (not (equal? next-type 'tab-panel))
               (not (and last? (equal? parent-type 'tab)))
          )
        result 
        #f
      )
    )
  )
  
  (provide hierarchy-up-func)
  (define (hierarchy-up-func button control-event)
    (let* 
      ((selected (send hierarchy-widget get-selected)))
      (if selected
        (send hierarchy-widget item-swap-up selected)
      )
    )
  )
  
  (provide hierarchy-up-one-func)
  (define (hierarchy-up-one-func button control-event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (send hierarchy-widget item-move-up selected move-up-cond)
      )
    )
  )
  
  (provide hierarchy-down-one-func)
  (define (hierarchy-down-one-func button control-event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (send hierarchy-widget item-move-down selected move-down-cond)
      )
    )
  )
  
  (provide hierarchy-down-func)
  (define (hierarchy-down-func button control-event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (send hierarchy-widget item-swap-down selected)
      )
    )
  )
  
  (provide hierarchy-delete-func)
  (define (hierarchy-delete-func button control-event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if (and selected (not (equal? selected "project")))
        ;; ... delete from the hierarchy list ...
        (send hierarchy-widget item-delete selected)
      )
    )
  )
  
  (provide hierarchy-hide-func)
  (define (hierarchy-hide-func button event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (let ((ptr (send hierarchy-widget item-property-get selected 'pointer #f)))
          (send hierarchy-widget item-property-set! selected 'hidden #t #f)
          (send ptr show #f)
          (update-hierarchy-buttons)
        )
      )
    )
  )
  
  (provide hierarchy-show-func)
  (define (hierarchy-show-func button event)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (let ((ptr (send hierarchy-widget item-property-get selected 'pointer #f)))
          (send hierarchy-widget item-property-set! selected 'hidden #f #f)
          (send ptr show #t)
          (update-hierarchy-buttons)
        )
      )
    )
  )
    
  ; callback function for the hierarchy widget
  (provide hierarchy-widget-func)
  ; action is a dotted pair
  ; car - the action, like select, collapse
  ; cdr - the id of the element for which the action took place
  (define (hierarchy-widget-func action id)
    ; ensure that the properties will be handed to the property window in order
    (define (sort-by-order-list lst order)
      (cond
        ((null? order) '())
        (else
         (cons (assoc (caar order) lst) (sort-by-order-list lst (cdr order)))
        )
      )
    )
        
    (cond
      ; select an item
      ; ------------------------------------------------------------------------
      ((eq? action 'select)
       (let*
         ((prop-list  (send hierarchy-widget get-item id))
          (type       (cadr (assoc 'type prop-list)))
          (order-list (widget-default-values type))
         )
         (send properties-panel import id type (sort-by-order-list prop-list order-list))
         ; set selection in the graph window as well
         (if (not (member type '(project frame)))
           (graph-widget-list-select id)
         )
         ; if it is a tab, switch in the tab panel
         (select-tab-above)
;;;         (if (equal? type 'tab)
;;;           (let
;;;             ((ptr (cadr (assoc 'pointer prop-list))))
;;;             (send ptr tab-select)
;;;           )
;;;         )
       )
      )
      ; clear the whole window
      ; ------------------------------------------------------------------------
      ((eq? action 'clear)
       (graph-widget-list-empty)
      )
      ; hide all windows before move up or down
      ; ------------------------------------------------------------------------
      ((member action '(before-move-up before-move-down))
       (preview-show-frames (send hierarchy-widget get-hierarchy) #f)
      )
      ; add an item
      ; ------------------------------------------------------------------------
      ((member action '(after-item-add after-item-append))
       (graph-widget-list-add id (send hierarchy-widget item-property-get id 'type #f))
      )
      ; delete an item
      ; ------------------------------------------------------------------------
      ((eq? action 'before-item-delete)
       (preview-show-frames (send hierarchy-widget get-hierarchy) #f)
       ;; ... delete from the graph window ...
       (graph-widget-list-del id)
      )
      ((eq? action 'after-item-delete)
       ;; ... delete the element in the preview ...
       (preview-update-all)
       ; ... show all windows ...
       (preview-show-frames (send hierarchy-widget get-hierarchy) #t)
       ;; ... empty out the property window ...
       (send properties-panel empty)
       ;; ... finally remember the change
       (changed! #f)
      )
      ; swap up or down an item
      ; ------------------------------------------------------------------------
      ((member action '(after-swap-down after-swap-up))
       ; update the preview
       ; this will only happen if not frames up swapped
       ; when frames are swapped, do nothing ...
       (preview-update-children-order id)
       ;; ... there is nothing to do for the property window
       ;; as the element is still selected ...
       ;; ... finally remember the change
       (changed! #f)
      )
      ; move up or down an item
      ; ------------------------------------------------------------------------
      ((member action '(after-move-up after-move-down))
       ; ... rebuild the preview ...
       (preview-update-all)
       ; ... show all windows ...
       (preview-show-frames (send hierarchy-widget get-hierarchy) #t)
       ;; ... finally remember the change
       (changed! #f)
      )
      ; a property has changed
      ; ------------------------------------------------------------------------
      ((equal? action 'before-item-property-set)
       (preview-show-frames (send hierarchy-widget get-hierarchy) #f)
      )
      ((equal? action 'after-item-property-set)
       ; ... rebuild the preview ...
       (preview-update-widget (car id) (cadr id) (caddr id))
       ; ... show all windows ...
       (preview-show-frames (send hierarchy-widget get-hierarchy) #t)
       ;; ... finally remember the change
       (changed! #f)
      )
      ; id change for an item
      ; ------------------------------------------------------------------------
      ((equal? action 'after-item-id-set!)
       (let*
         ((old-id (car id))
          (new-id (cadr id))
          (ptr    (send hierarchy-widget item-property-get new-id 'pointer #f))
          (type   (send hierarchy-widget item-property-get new-id 'type #f))
         )
         ; ... update the preview ...
         (send ptr mred-id-set! new-id)
         ; ... update the graph editor ...
         (graph-widget-list-change-id type old-id new-id)
         ;; ... finally remember the change
         (changed! #f)
       )
      )
    )
    
    ; update buttons in the hierarchy widget
    (update-hierarchy-buttons)
  )
  
  (provide hierarchy-widget-key-func)
  (define (hierarchy-widget-key-func key shift? ctrl?)
    (let ((selected (send hierarchy-widget get-selected)))
      (cond
        ((equal? key #\rubout)
         (if (and selected (not (equal? selected "project")))
           ;; ... delete from the hierarchy list ...
           (send hierarchy-widget item-delete selected)
         )
        )
        ((and (equal? key 'up) shift?)
         (if selected
           (send hierarchy-widget item-swap-up selected)
         )
        )
        ((and (equal? key 'down) shift?)
         (if selected
           (send hierarchy-widget item-swap-down selected)
         )
        )
        ((and ctrl? (equal? key #\c))
         (if selected
           (clipboard-copy selected #t)
         )
        )
        ((and ctrl? (equal? key #\x))
         (if selected
           (clipboard-cut selected)
         )
        )
        ((and ctrl? (equal? key #\v))
         (if selected
           (clipboard-paste selected)
         )
        )
      )
    )
  )
  
  
  ;; This function updates the enabled/disabled buttons status.
  (provide update-hierarchy-buttons)
  (define (update-hierarchy-buttons)
    (let ((selected (send hierarchy-widget get-selected)))
      (if selected
        (let ((type (send hierarchy-widget item-property-get selected 'type #f)))
          ; if a frame (window) is selected
          (if (eq? type 'frame)
            ;; Ok this is a frame. Is it shown?
            (if (send (send hierarchy-widget item-property-get selected 'pointer #f) is-shown?)
              (begin
                (send hierarchy-show-button enable #f)
                (send hierarchy-hide-button enable #t))
              (begin 
                (send hierarchy-show-button enable #t)
                (send hierarchy-hide-button enable #f)))
            ;; This is not a frame. Therefore disable both buttons.
            (begin
              (send hierarchy-show-button enable #f)
              (send hierarchy-hide-button enable #f)))
          ; if the project is selected
          (if (eq? type 'project)
            (send hierarchy-delete-button enable #f)
            (send hierarchy-delete-button enable #t))
          ; check swapping states
          (if (send hierarchy-widget item-can-swap-up? selected)
            (send hierarchy-up-button enable #t)
            (send hierarchy-up-button enable #f))
          (if (send hierarchy-widget item-can-move-up? selected move-up-cond)
            (send hierarchy-up-one-button enable #t)
            (send hierarchy-up-one-button enable #f))
          (if (send hierarchy-widget item-can-move-down? selected move-down-cond)
            (send hierarchy-down-one-button enable #t)
            (send hierarchy-down-one-button enable #f))
          (if (send hierarchy-widget item-can-swap-down? selected)
            (send hierarchy-down-button enable #t)
            (send hierarchy-down-button enable #f)))
        (begin
          (send hierarchy-up-button enable #f)
          (send hierarchy-up-one-button enable #f)
          (send hierarchy-down-one-button enable #f)
          (send hierarchy-down-button enable #f)
          (send hierarchy-delete-button enable #f)
          (send hierarchy-show-button enable #f)
          (send hierarchy-hide-button enable #f)))))

  ; -----------------------------------------------------------------------------
  ; clipboard
  ; -----------------------------------------------------------------------------
    
    ; this is a clipboard area when a widget is copied
    ; this is #f or an atom or a list of atoms
    (define item-clipboard #f)
    ; this is a list of properties which corresponds to an atom or
    ; to the list
    (define item-clipboard-prop #f)
    
    ; determines whether there is anything in the clipboard
    (provide clipboard?)
    (define (clipboard?)
      (if item-clipboard #t #f)
    )
    
    ; it copies only one widget
    (define (widget-copy selected)
      (let
        ((props (send hierarchy-widget get-item selected)))
        (set! item-clipboard-prop props)
        (set! item-clipboard selected)))
            
    ; it copies a container and all of its children
    (define (container-copy selected)
      ; it finds the id and if it is a container then considers it children as well
      ; it returns 
      ; - #f : this is actually an error
      ; - a list : the container and all of its children
      ; - an atom : just a widget
      (define (item-find id lst)
        (cond
          ((null? lst) #f)
          ((list? (car lst))
           (if (equal? (caar lst) id)
             (car lst)
             (let
               ((found? (item-find id (cdar lst))))
               (if found? found? (item-find id (cdr lst))))))
          (else
           (if (equal? (car lst) id)
             (car lst)
             (item-find id (cdr lst))))))
      
      ; provided a list of widget or a widget, 
      ; it copies its data according to the function
      (define (container-copy-aux lst)
        (cond
          ((null? lst) '())
          ((and (list? lst) (list? (car lst)))
           (cons (container-copy-aux (car lst)) 
                 (container-copy-aux (cdr lst)))
          )
          ((list? lst)
           (let
             ((props (send hierarchy-widget get-item (car lst))))
             (cons props (container-copy-aux (cdr lst)))
           )
          )
          (else
           (send hierarchy-widget get-item lst)
          )
        )
      )
      
      (set! item-clipboard (item-find selected (send hierarchy-widget get-hierarchy)))
      (if item-clipboard
        (set! item-clipboard-prop (container-copy-aux item-clipboard))
      )
    )
    
    ; public interface for copying a widget
    (provide clipboard-copy)
    (define (clipboard-copy id with-children?)
      (if id
        (begin
          (if with-children?
            (container-copy id)
            (widget-copy id)
          )
        )
      )
    )
    
    ; public interface for cutting a widget
    ; this includes a copy and a delete operation
    (provide clipboard-cut)
    (define (clipboard-cut id)
      (if (and id (not (equal? id "project")))
        (begin
          (clipboard-copy id #t)
          (send hierarchy-widget item-delete id)
        )
      )
    )
    
    (provide clipboard-paste)
    (define (clipboard-paste id)
      
      ; paste one item
      (define (item-paste parent-id props container? after?)
        (let* 
          ((item-props (assoc-remove props '(collapsed icon pointer)))
           ; determine the type
           (type (cadr (assoc 'type props)))
           ; generate an id symbol
           (id-sym (gensym (string->symbol (string-append (symbol->string type) "-"))))
           ; an id string
           (id-str (symbol->string id-sym))
          )
          
          (send hierarchy-widget 
                item-add
                parent-id 
                id-str
                (cons '(pointer #f) item-props)
                after? 
                (icon-get type)
                container?
                #f)
          
          ; return the string id
          id-str
        )
      )
      
      ; paste a container and its contents
      (define (container-paste parent lst lst-prop after?)
        (cond
          ((null? lst) #f)
          ((list? (car lst))
           (let*
             ((id (item-paste parent (caar lst-prop) #t after?)))
             (container-paste id (cdar lst) (cdar lst-prop) #f) ; rest of the container
             (container-paste id (cdr lst) (cdr lst-prop) #t)
           )
          )
          (else
           (let
             ((id (item-paste parent (car lst-prop) #f after?)))
             (container-paste id (cdr lst) (cdr lst-prop) #f)
           )
          )
        )
      )
      
      (let*
        ; determine the type of the element in the clipboard
        ((head (car item-clipboard-prop))
         (head-type (cadr (assoc 'type head)))
        )
        (if (or (and (not (equal? id "project"))
                     (or (not (list? item-clipboard))
                         (and (list? item-clipboard)
                              (not (equal? head-type 'frame)))))
                (and (equal? id "project")
                     (list? item-clipboard)
                     (equal? head-type 'frame))
            )
          (let 
            ((old-lst (send hierarchy-widget get-hierarchy)))
            ; hide all frames
            ; we have to do this, 
            ; as we may insert the clipboard to anywhere in the window
            ; and MrEd can add the widget only to the end
            (preview-show-frames old-lst #f)
            (send hierarchy-widget update-view! #f)
            ; paste the clipboard
            (cond 
              ((list? item-clipboard)
               (container-paste id (list item-clipboard) (list item-clipboard-prop) #f)
              )
              (item-clipboard
               (item-paste id item-clipboard-prop #f #f)
              )
            )
            (send hierarchy-widget update-view! #t)
            (changed! #f)
            ; rebuild the windows
            (preview-update-all)
            ; show the windows
            (preview-show-frames (send hierarchy-widget get-hierarchy) #t)
          )
        )
      )
    )

  
) ; end of module

